# coding: utf-8
import logging
import os
import uuid
from concurrent import futures

import grpc

from taichu_serve.grpc_predict_v2_pb2_grpc import add_GRPCInferenceServiceServicer_to_server
from taichu_serve.app import app, init_model_service_instance
from taichu_serve.grpc_server import GrpcModelService, GrpcServerInterceptor
from taichu_serve.settings import parse_args
from taichu_serve.dockerfile import Dockerfile_template, launch_sh_template

LOGGER = logging.getLogger(__name__)


def init_project(name="project"):
    # 在当前目录下创建项目目录
    if not os.path.exists(name):
        os.mkdir(name)

    # 获取当前目录
    current_path = os.path.abspath(os.path.dirname(__file__))
    # 获取模板目录
    template_path = os.path.join(current_path, "template")
    # 拷贝模板目录下的文件到项目目录
    for file in os.listdir(template_path):
        if file == "__pycache__":
            continue
        if file.endswith("__init__.py"):
            continue

        if os.path.isdir(os.path.join(template_path, file)):
            os.mkdir(os.path.join(name, file))
            for sub_file in os.listdir(os.path.join(template_path, file)):
                if sub_file == "__pycache__":
                    continue
                if sub_file.endswith("__init__.py"):
                    continue
                with open(os.path.join(template_path, file, sub_file), "r") as f:
                    content = f.read()
                print("create file: ", os.path.join(name, file, sub_file))
                with open(os.path.join(name, file, sub_file), "w") as f:
                    f.write(content)
            continue

        with open(os.path.join(template_path, file), "r") as f:
            content = f.read()
        print("create file: ", os.path.join(name, file))
        with open(os.path.join(name, file), "w") as f:
            f.write(content)

    with open(os.path.join(name, "dependencies.txt"), "w") as f:
        f.write("# 请在这里添加apt-get安装的依赖，每行一个依赖，如：curl")
    print("create file: ", os.path.join(name, "dependencies.txt"))
    print("init project done!")


def build_env(base_image):
    # 1. 创建dockerfile
    if base_image is None or base_image == "":
        base_image = "swr.cn-central-221.ovaijisuan.com/wair/taichu-serve:latest"
    if os.path.exists("Dockerfile"):
        print("Dockerfile已存在，请删除后重试！")
        return

    if os.path.exists("launch.sh"):
        os.remove("launch.sh")

    if not os.path.exists("requirements.txt"):
        with open("requirements.txt", "w") as f:
            f.write("taichu-serve")
        print("create file: ", "requirements.txt")

    if not os.path.exists("dependencies.txt"):
        with open("dependencies.txt", "w") as f:
            f.write("# 请在这里添加apt-get安装的依赖，每行一个依赖，如：curl")
        print("create file: ", "dependencies.txt")

    # 逐行读取dependencies.txt
    with open("dependencies.txt", "r") as f:
        dependencies = f.readlines()
        dependencies = [x.strip() for x in dependencies]
        # 跳过注释行
        dependencies = [x for x in dependencies if not x.startswith("#")]
        dependencies = [x for x in dependencies if x != ""]

    apt_install = ''
    for dependency in dependencies:
        if not dependency.startswith("RUN apt-get install -y"):
            apt_install = 'RUN  apt-get update && apt-get install -y '
        apt_install += dependency + " "
    dockerfile_content = Dockerfile_template.format(base_image=base_image, apt_install=apt_install)

    with open("launch.sh", "w") as f:
        f.write(launch_sh_template)
    print("create file: ", "launch.sh")
    with open("Dockerfile", "w") as f:
        f.write(dockerfile_content)
    print("create file: ", "Dockerfile")

    print("build docker image...")
    tag = uuid.uuid4().hex[:8]
    print("[CMD] docker build -t taichu-serve-env:{tag} .".format(tag=tag))
    cmd = "docker build -t taichu-serve-env:{tag} .".format(tag=tag)
    os.system(cmd)
    print("build docker image done!")

    print("进入镜像调试命令：")
    print("[CMD] docker run -it -p 8080:8080 -p 8081:8081 -v $PWD:/home taichu-serve-env:{tag} bash".format(tag=tag))

    print("调试完成后，可以使用以下命令将镜像推送到仓库：")
    image_name = "taichu-serve-env:{tag}".format(tag=tag)
    print("[CMD] docker tag {image_name} swr.cn-central-221.ovaijisuan.com/wair/{image_name}".format(image_name=image_name))
    print("[CMD] docker push swr.cn-central-221.ovaijisuan.com/wair/{image_name}".format(image_name=image_name))


def cli():
    args = parse_args()
    if args.action == "init":
        init_project()
        return
    if args.action == "build_env":
        build_env(args.base_image)
        return

    init_model_service_instance()
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=args.max_concurrent_requests),
                         maximum_concurrent_rpcs=args.max_concurrent_requests + 5,
                         interceptors=[GrpcServerInterceptor()])

    add_GRPCInferenceServiceServicer_to_server(GrpcModelService(), server)
    server.add_insecure_port(f'[::]:{args.grpc_port}')

    server.start()
    LOGGER.info("grpc server start at port %s", args.grpc_port)

    if args.grpc_only:
        server.wait_for_termination()

    app.run("0.0.0.0", args.http_port)
    return app
